"""
This module contains the specific code for generating standalone or airgapped ISOs.
"""

# SPDX-License-Identifier: GPL-2.0-or-later

import os
import re
from typing import Dict, List

from cobbler import utils

from cobbler.actions import buildiso

cdregex = re.compile(r"^\s*url .*\n", re.IGNORECASE | re.MULTILINE)


def _generate_append_line_standalone(data: dict, distro, descendant) -> str:
    """
    Generates the append line for the kernel so the installation can be done unattended.
    :param data: The values for the append line. The key "kernel_options" must be present.
    :param distro: The distro object to generate the append line from.
    :param descendant: The profile or system which is underneath the distro.
    :return: The base append_line which we need for booting the built ISO. Contains initrd and autoinstall parameter.
    """
    append_line = "  append initrd=%s" % os.path.basename(distro.initrd)
    if distro.breed == "redhat":
        append_line += " inst.ks=cdrom:/isolinux/%s.cfg" % descendant.name
    elif distro.breed == "suse":
        append_line += (
            " autoyast=file:///isolinux/%s.cfg install=cdrom:///" % descendant.name
        )
        if "install" in data["kernel_options"]:
            del data["kernel_options"]["install"]
    elif distro.breed in ["ubuntu", "debian"]:
        append_line += (
            " auto-install/enable=true preseed/file=/cdrom/isolinux/%s.cfg"
            % descendant.name
        )

    # add remaining kernel_options to append_line
    append_line += buildiso.add_remaining_kopts(data["kernel_options"])
    return append_line


class StandaloneBuildiso(buildiso.BuildIso):
    """
    This class contains all functionality related to building self-contained installation images.
    """

    def _validate_standalone_args(self, distro_name: str, source: str):
        """
        If building standalone, we only want --distro and --profiles (optional), systems are disallowed
        :param distro_name: The name of the distribution we want to install.
        :param source: The source which we copy to the ISO as an installation base.
        """
        if not distro_name:
            raise ValueError(
                "When building a standalone ISO, you must specify a --distro"
            )
        base_distro = self.api.find_distro(name=distro_name)
        if base_distro is None:
            raise ValueError("Distro specified was not found!")
        source = self._validate_standalone_filesource(source, base_distro)
        if not os.path.exists(source):
            raise ValueError('The specified source "%s" does not exist!' % source)

        # Ensure all profiles specified are children of the distro
        if self.profiles:
            orphan_profiles = [
                profile
                for profile in self.profiles
                if profile not in base_distro.children
            ]
            if len(orphan_profiles) > 0:
                raise ValueError(
                    "When building a standalone ISO, all --profiles must be under --distro"
                )

    def _sync_airgapped_repos(self, airgapped: bool, repo_names_to_copy: dict):
        """
        Syncs all repositories locally available into the image if the is supposed to be airgapped.
        :param airgapped: If false this method is doing nothing.
        :param repo_names_to_copy: The names of the repositories which should be included.
        :raises RuntimeError: In case the rsync of the repository was not successful.
        """
        if airgapped:
            # copy any repos found in profiles or systems to the iso build
            repodir = os.path.abspath(
                os.path.join(self.isolinuxdir, "..", "repo_mirror")
            )
            if not os.path.exists(repodir):
                os.makedirs(repodir)

            for repo_name in repo_names_to_copy:
                self.logger.info(" - copying repo '%s' for airgapped ISO", repo_name)
                rsync_successful = utils.rsync_files(
                    repo_names_to_copy[repo_name],
                    os.path.join(repodir, repo_name),
                    "--exclude=TRANS.TBL --exclude=cache/ --no-g",
                    quiet=True,
                )
                if not rsync_successful:
                    raise RuntimeError('rsync of repo "%s" failed' % repo_name)

    def _validate_standalone_filesource(self, filesource: str, distro) -> str:
        """
        Validate that the path to the installation sources is making sense. If they do then normalize the path and
        return it.
        :param filesource: The path to the installation sources.
        :param distro: The distribution of which the kernel is used for booting the image.
        :raises ValueError: In case the installation source was not found.
        :return: Normalized filesource which points to the absolute path of the source tree.
        """
        if not filesource:
            # Try to determine the source from the distro kernel path
            self.logger.debug("Trying to locate source for distro")
            (source_head, source_tail) = os.path.split(distro.kernel)
            distro_mirror = os.path.join(self.api.settings().webdir, "distro_mirror")
            while source_tail != "":
                if source_head == distro_mirror:
                    filesource = os.path.join(source_head, source_tail)
                    self.logger.debug("Found source in %s", filesource)
                    return filesource
                (source_head, source_tail) = os.path.split(source_head)
            # Can't find the source, raise an error
            raise ValueError(
                "Error, no installation source found. When building a standalone ISO, you must specify a "
                "--source if the distro install tree is not hosted locally"
            )
        return filesource

    def _generate_autoinstall_data(
        self, descendant, distro, airgapped: bool, data: dict, repo_names_to_copy: dict
    ) -> str:
        """
        Generates the autoinstall script for the distro/profile/system we want to install.
        :param descendant: The descendant to generate the ISO for.
        :param distro: The distro to generate the ISO for.
        :param airgapped: Whether the ISO should be bootable in an airgapped environment or not.
        :param data: The data for the append line of the kernel.
        :param repo_names_to_copy: The repositories to include in case of an airgapped environment.
        :return: The generated script for the ISO.
        """
        autoinstall_data = ""
        if descendant.COLLECTION_TYPE == "profile":
            autoinstall_data = self.api.autoinstallgen.generate_autoinstall_for_profile(
                descendant.name
            )
        elif descendant.COLLECTION_TYPE == "system":
            autoinstall_data = self.api.autoinstallgen.generate_autoinstall_for_system(
                descendant.name
            )

        if distro.breed == "redhat":
            autoinstall_data = cdregex.sub("cdrom\n", autoinstall_data, count=1)

        if airgapped:
            for repo_name in data.get("repos", []):
                repo_obj = self.api.find_repo(repo_name)
                error = (
                    "%s %s refers to repo %s, which {error_message}; cannot build airgapped ISO"
                    % (descendant.COLLECTION_TYPE, descendant.name, repo_name)
                )

                if repo_obj is None:
                    raise ValueError(error.format(error_message="does not exist"))
                if not repo_obj.mirror_locally:
                    raise ValueError(
                        error.format(
                            error_message="is not configured for local mirroring"
                        )
                    )
                mirrordir = os.path.join(
                    self.api.settings().webdir, "repo_mirror", repo_obj.name
                )
                if not os.path.exists(mirrordir):
                    raise ValueError(
                        error.format(
                            error_message="has a missing local mirror directory"
                        )
                    )

                repo_names_to_copy[repo_obj.name] = mirrordir

                # update the baseurl in autoinstall_data to use the cdrom copy of this repo
                reporegex = re.compile(
                    r"^(\s*repo --name=%s --baseurl=).*" % repo_obj.name, re.MULTILINE
                )
                autoinstall_data = reporegex.sub(
                    r"\1" + "file:///mnt/source/repo_mirror/" + repo_obj.name,
                    autoinstall_data,
                )

            # rewrite any split-tree repos, such as in redhat, to use cdrom
            srcreporegex = re.compile(
                r"^(\s*repo --name=\S+ --baseurl=).*/cobbler/distro_mirror/%s/?(.*)"
                % distro.name,
                re.MULTILINE,
            )
            autoinstall_data = srcreporegex.sub(
                r"\1" + "file:///mnt/source" + r"\2", autoinstall_data
            )
        return autoinstall_data

    def _generate_descendant(
        self,
        descendant,
        cfglines: List[str],
        distro,
        airgapped: bool,
        repo_names_to_copy: dict,
    ):
        """
        Generate the ISOLINUX cfg configuration file for the descendant.
        :param descendant: The descendant to generate the config file for. Must be a profile or system object.
        :param cfglines: The content of the file which has already been generated.
        :param distro: The parent distro.
        :param airgapped: Whether the generated ISO should be bootable in an airgapped environment or not.
        :param repo_names_to_copy: The repository names to copy in the case of an airgapped environment.
        """
        menu_indent = 0
        if descendant.COLLECTION_TYPE == "system":
            menu_indent = 4

        data = utils.blender(self.api, False, descendant)

        # SUSE is not using 'text'. Instead 'textmode' is used as kernel option.
        if distro is not None:
            utils.kopts_overwrite(
                data["kernel_options"], self.api.settings().server, distro.breed
            )

        cfglines.append("")
        cfglines.append("LABEL %s" % descendant.name)
        if menu_indent:
            cfglines.append("  MENU INDENT %d" % menu_indent)
        cfglines.append("  MENU LABEL %s" % descendant.name)
        cfglines.append("  kernel %s" % os.path.basename(distro.kernel))

        cfglines.append(_generate_append_line_standalone(data, distro, descendant))

        autoinstall_data = self._generate_autoinstall_data(
            descendant, distro, airgapped, data, repo_names_to_copy
        )
        autoinstall_name = os.path.join(self.isolinuxdir, "%s.cfg" % descendant.name)
        with open(autoinstall_name, "w+") as autoinstall_file:
            autoinstall_file.write(autoinstall_data)

    def generate_standalone_iso(
        self, distro_name: str, filesource: str, airgapped: bool
    ):
        """Creates the ``isolinux.cfg`` for a standalone or airgapped ISO image. And copies possible repositories to
        the image source folder.
        :param distro_name: The name of the Cobbler distribution.
        :param filesource: The source directory for the ISO. This gets rsynced into isolinuxdir.
        :param airgapped: Whether the repositories have to be locally available or the internet is reachable.
        """
        # Get the distro object for the requested distro and then get all of its descendants (profiles/sub-profiles/
        # systems) with sort=True for profile/system hierarchy to allow menu indenting
        distro = self.api.find_distro(name=distro_name)
        if distro is None:
            raise ValueError(
                'Distro "%s" was not found, aborting generation of ISO-file!'
                % distro_name
            )

        self.copy_boot_files(distro, self.isolinuxdir)

        self.logger.info("generating an isolinux.cfg")

        cfglines = [self.iso_template]

        repo_names_to_copy: Dict[str, str] = {}

        for descendant in distro.children:
            descendant = self.api.find_items(what="", name=descendant)
            # if a list of profiles was given, skip any others and their systems
            if len(self.profiles) > 0 and (
                (
                    descendant.COLLECTION_TYPE == "profile"
                    and descendant.name not in self.profiles
                )
                or (
                    descendant.COLLECTION_TYPE == "system"
                    and descendant.profile not in self.profiles
                )
            ):
                continue
            self._generate_descendant(
                descendant, cfglines, distro, airgapped, repo_names_to_copy
            )

        self.logger.info("done writing config")
        cfglines.append("")
        cfglines.append("MENU END")
        with open(os.path.join(self.isolinuxdir, "isolinux.cfg"), "w+") as cfg:
            cfg.writelines(cfglines)

        self._sync_airgapped_repos(airgapped, repo_names_to_copy)

        # copy distro files last, since they take the most time
        cmd = [
            "rsync",
            "-rlptgu",
            "--exclude",
            "boot.cat",
            "--exclude",
            "TRANS.TBL",
            "--exclude",
            "isolinux/",
            "%s/" % filesource,
            "%s/../" % self.isolinuxdir,
        ]
        self.logger.info('- copying distro "%s" files (%s)', distro_name, cmd)
        rsync_return_code = utils.subprocess_call(cmd, shell=False)
        if rsync_return_code:
            raise OSError("rsync of distro files failed")

    def run(
        self,
        iso: str = "autoinst.iso",
        buildisodir: str = "",
        profiles: List[str] = None,
        xorrisofs_opts: str = "",
        distro_name: str = "",
        airgapped: bool = False,
        source="",
    ):
        """
        Run the whole iso generation from bottom to top. Per default this builds an ISO for all available systems
        and profiles.
        This is the only method which should be called from non-class members. The ``profiles`` and ``system``
        parameters can be combined.
        :param iso: The name of the iso. Defaults to "autoinst.iso".
        :param buildisodir: This overwrites the directory from the settings in which the iso is built in.
        :param profiles: The filter to generate the ISO only for selected profiles.
        :param xorrisofs_opts: ``xorrisofs`` options to include additionally.
        :param distro_name: For detecting the architecture of the ISO.
        :param airgapped: This option implies ``standalone=True``.
        :param source: If the iso should be offline available this is the path to the sources of the image.
        """
        buildisodir = self._prepare_iso(buildisodir, distro_name, profiles)
        self._validate_standalone_args(distro_name, source)
        self.generate_standalone_iso(distro_name, source, airgapped)
        self._generate_iso(xorrisofs_opts, iso, buildisodir)



"""
This module is responsible for containing all enums we use in Cobbler. It should not be dependent upon any other module
except the Python standard library.
"""

import enum
from typing import Union

VALUE_INHERITED = "<<inherit>>"
VALUE_NONE = "none"


class ConvertableEnum(enum.Enum):
    """
    Abstract class to convert the enum via our convert method.
    """

    @classmethod
    def to_enum(cls, value: Union[str, "ConvertableEnum"]) -> "ConvertableEnum":
        """
        This method converts the chosen str to the corresponding enum type.

        :param value: str which contains the to be converted value.
        :returns: The enum value.
        :raises TypeError: In case value was not of type str.
        :raises ValueError: In case value was not in the range of valid values.
        """
        try:
            if isinstance(value, str):
                if value == VALUE_INHERITED:
                    try:
                        return cls["INHERITED"]
                    except KeyError as key_error:
                        raise ValueError("The enum given does not support inheritance!") from key_error
                return cls[value.upper()]
            elif isinstance(value, cls):
                return value
            else:
                raise TypeError(f"{value} must be a str or Enum")
        except KeyError:
            raise ValueError(f"{value} must be one of {list(cls)}")


class ResourceAction(ConvertableEnum):
    """
    This enum represents all actions a resource may execute.
    """

    CREATE = "create"
    REMOVE = "remove"


class NetworkInterfaceType(enum.Enum):
    """
    This enum represents all interface types Cobbler is able to set up on a target host.
    """

    NA = 0
    BOND = 1
    BOND_SLAVE = 2
    BRIDGE = 3
    BRIDGE_SLAVE = 4
    BONDED_BRIDGE_SLAVE = 5
    BMC = 6
    INFINIBAND = 7


class RepoBreeds(ConvertableEnum):
    """
    This enum describes all repository breeds Cobbler is able to manage.
    """

    NONE = VALUE_NONE
    RSYNC = "rsync"
    RHN = "rhn"
    YUM = "yum"
    APT = "apt"
    WGET = "wget"


class RepoArchs(ConvertableEnum):
    """
    This enum describes all repository architectures Cobbler is able to serve in case the content of the repository is
    serving the same architecture.
    """

    NONE = VALUE_NONE
    I386 = "i386"
    X86_64 = "x86_64"
    IA64 = "ia64"
    PPC = "ppc"
    PPC64 = "ppc64"
    PPC64LE = "ppc64le"
    PPC64EL = "ppc64el"
    S390 = "s390"
    ARM = "arm"
    AARCH64 = "aarch64"
    NOARCH = "noarch"
    SRC = "src"


class Archs(ConvertableEnum):
    """
    This enum describes all system architectures which Cobbler is able to provision.
    """

    I386 = "i386"
    X86_64 = "x86_64"
    IA64 = "ia64"
    PPC = "ppc"
    PPC64 = "ppc64"
    PPC64LE = "ppc64le"
    PPC64EL = "ppc64el"
    S390 = "s390"
    S390X = "s390x"
    ARM = "arm"
    AARCH64 = "aarch64"


class VirtType(ConvertableEnum):
    """
    This enum represents all known types of virtualization Cobbler is able to handle via Koan.
    """

    INHERITED = VALUE_INHERITED
    QEMU = "qemu"
    KVM = "kvm"
    XENPV = "xenpv"
    XENFV = "xenfv"
    VMWARE = "vmware"
    VMWAREW = "vmwarew"
    OPENVZ = "openvz"
    AUTO = "auto"


class VirtDiskDrivers(ConvertableEnum):
    """
    This enum represents all virtual disk driver Cobbler can handle.
    """

    INHERITED = VALUE_INHERITED
    RAW = "raw"
    QCOW2 = "qcow2"
    QED = "qed"
    VDI = "vdi"
    VDMK = "vdmk"


class BaudRates(enum.Enum):
    """
    This enum describes all baud rates which are commonly used.
    """

    DISABLED = -1
    B0 = 0
    B110 = 110
    B300 = 300
    B600 = 600
    B1200 = 1200
    B2400 = 2400
    B4800 = 4800
    B9600 = 9600
    B14400 = 14400
    B19200 = 19200
    B38400 = 38400
    B57600 = 57600
    B115200 = 115200
    B128000 = 128000
    B256000 = 256000


class ImageTypes(ConvertableEnum):
    """
    This enum represents all image types which Cobbler can manage.
    """

    DIRECT = "direct"
    ISO = "iso"
    MEMDISK = "memdisk"
    VIRT_CLONE = "virt-clone"


class MirrorType(ConvertableEnum):
    """
    This enum represents all mirror types which Cobbler can manage.
    """

    METALINK = "metalink"
    MIRRORLIST = "mirrorlist"
    BASEURL = "baseurl"


class TlsRequireCert(ConvertableEnum):
    """
    This enum represents all TLS validation server cert types which Cobbler can manage.
    """

    NEVER = "never"
    ALLOW = "allow"
    DEMAND = "demand"
    HARD = "hard"



"""
This module contains the specific code to generate a network bootable ISO.
"""

# SPDX-License-Identifier: GPL-2.0-or-later

import os
import re
from typing import List

from cobbler import utils
from cobbler.actions import buildiso


class AppendLineBuilder:
    """
    This class is meant to be initiated for a single append line. Afterwards the object should be disposed.
    """

    def __init__(self, distro_name: str, data: dict):
        self.append_line = ""
        self.data = data
        self.distro_name = distro_name
        self.dist = None
        self.system_interface = None
        self.system_ip = None
        self.system_netmask = None
        self.system_gw = None
        self.system_dns = None

    def _system_int_append_line(self):
        """
        This generates the interface configuration for the system to boot for the append line.
        """
        if self.system_interface is not None:
            intmac = "mac_address_" + self.system_interface
            if self.dist.breed == "suse":
                if self.data.get(intmac, "") != "":
                    self.append_line += (
                        " netdevice=%s"
                        % self.data["mac_address_" + self.system_interface].lower()
                    )
                else:
                    self.append_line += " netdevice=%s" % self.system_interface
            elif self.dist.breed == "redhat":
                if self.data.get(intmac, "") != "":
                    self.append_line += (
                        " ksdevice=%s"
                        % self.data["mac_address_" + self.system_interface]
                    )
                else:
                    self.append_line += " ksdevice=%s" % self.system_interface
            elif self.dist.breed in ["ubuntu", "debian"]:
                self.append_line += (
                    " netcfg/choose_interface=%s" % self.system_interface
                )

    def _system_ip_append_line(self):
        """
        This generates the IP configuration for the system to boot for the append line.
        """
        if self.system_ip is not None:
            if self.dist.breed == "suse":
                self.append_line += " hostip=%s" % self.system_ip
            elif self.dist.breed == "redhat":
                self.append_line += " ip=%s" % self.system_ip
            elif self.dist.breed in ["ubuntu", "debian"]:
                self.append_line += " netcfg/get_ipaddress=%s" % self.system_ip

    def _system_mask_append_line(self):
        """
        This generates the netmask configuration for the system to boot for the append line.
        """
        if self.system_netmask is not None:
            if self.dist.breed in ["suse", "redhat"]:
                self.append_line += " netmask=%s" % self.system_netmask
            elif self.dist.breed in ["ubuntu", "debian"]:
                self.append_line += " netcfg/get_netmask=%s" % self.system_netmask

    def _system_gw_append_line(self):
        """
        This generates the gateway configuration for the system to boot for the append line.
        """
        if self.system_gw is not None:
            if self.dist.breed in ["suse", "redhat"]:
                self.append_line += " gateway=%s" % self.system_gw
            elif self.dist.breed in ["ubuntu", "debian"]:
                self.append_line += " netcfg/get_gateway=%s" % self.system_gw

    def _system_dns_append_line(self, exclude_dns: bool):
        """
        This generates the DNS configuration for the system to boot for the append line.
        :param exclude_dns: If this flag is set to True, the DNS configuration is skipped.
        """
        if not exclude_dns or self.system_dns is not None:
            if self.dist.breed == "suse":
                nameserver_key = "nameserver"
            elif self.dist.breed == "redhat":
                nameserver_key = "dns"
            elif self.dist.breed in ["ubuntu", "debian"]:
                nameserver_key = "netcfg/get_nameservers"
            else:
                return

            if isinstance(self.system_dns, list):
                joined_nameservers = ",".join(self.system_dns)
                if joined_nameservers != "":
                    self.append_line += " %s=%s" % (nameserver_key, joined_nameservers)
            else:
                self.append_line += " %s=%s" % (nameserver_key, self.system_dns)

    def _generate_static_ip_boot_interface(self):
        """
        The interface to use when the system boots.
        """
        if self.dist.breed == "redhat":
            if self.data["kernel_options"].get("ksdevice", "") != "":
                self.system_interface = self.data["kernel_options"]["ksdevice"]
                if self.system_interface == "bootif":
                    self.system_interface = None
                del self.data["kernel_options"]["ksdevice"]
        elif self.dist.breed == "suse":
            if self.data["kernel_options"].get("netdevice", "") != "":
                self.system_interface = self.data["kernel_options"]["netdevice"]
                del self.data["kernel_options"]["netdevice"]
        elif self.dist.breed in ["debian", "ubuntu"]:
            if self.data["kernel_options"].get("netcfg/choose_interface", "") != "":
                self.system_interface = self.data["kernel_options"][
                    "netcfg/choose_interface"
                ]
                del self.data["kernel_options"]["netcfg/choose_interface"]

    def _generate_static_ip_boot_ip(self):
        """
        Generate the IP which is used during the installation process. This respects overrides.
        """
        if self.dist.breed == "redhat":
            if self.data["kernel_options"].get("ip", "") != "":
                self.system_ip = self.data["kernel_options"]["ip"]
                del self.data["kernel_options"]["ip"]
        elif self.dist.breed == "suse":
            if self.data["kernel_options"].get("hostip", "") != "":
                self.system_ip = self.data["kernel_options"]["hostip"]
                del self.data["kernel_options"]["hostip"]
        elif self.dist.breed in ["debian", "ubuntu"]:
            if self.data["kernel_options"].get("netcfg/get_ipaddress", "") != "":
                self.system_ip = self.data["kernel_options"]["netcfg/get_ipaddress"]
                del self.data["kernel_options"]["netcfg/get_ipaddress"]

    def _generate_static_ip_boot_mask(self):
        """
        Generate the Netmask which is used during the installation process. This respects overrides.
        """
        if self.dist.breed in ["suse", "redhat"]:
            if self.data["kernel_options"].get("netmask", "") != "":
                self.system_netmask = self.data["kernel_options"]["netmask"]
                del self.data["kernel_options"]["netmask"]
        elif self.dist.breed in ["debian", "ubuntu"]:
            if self.data["kernel_options"].get("netcfg/get_netmask", "") != "":
                self.system_netmask = self.data["kernel_options"]["netcfg/get_netmask"]
                del self.data["kernel_options"]["netcfg/get_netmask"]

    def _generate_static_ip_boot_gateway(self):
        """
        Generate the Gateway which is used during the installation process. This respects overrides.
        """
        if self.dist.breed in ["suse", "redhat"]:
            if self.data["kernel_options"].get("gateway", "") != "":
                self.system_gw = self.data["kernel_options"]["gateway"]
                del self.data["kernel_options"]["gateway"]
        elif self.dist.breed in ["debian", "ubuntu"]:
            if self.data["kernel_options"].get("netcfg/get_gateway", "") != "":
                self.system_gw = self.data["kernel_options"]["netcfg/get_gateway"]
                del self.data["kernel_options"]["netcfg/get_gateway"]

    def _generate_static_ip_boot_dns(self):
        """
        Generates the static Boot DNS Server which is used for resolving Domains.
        """
        if self.dist.breed == "redhat":
            if self.data["kernel_options"].get("dns", "") != "":
                self.system_dns = self.data["kernel_options"]["dns"]
                del self.data["kernel_options"]["dns"]
        elif self.dist.breed == "suse":
            if self.data["kernel_options"].get("nameserver", "") != "":
                self.system_dns = self.data["kernel_options"]["nameserver"]
                del self.data["kernel_options"]["nameserver"]
        elif self.dist.breed in ["debian", "ubuntu"]:
            if self.data["kernel_options"].get("netcfg/get_nameservers", "") != "":
                self.system_dns = self.data["kernel_options"]["netcfg/get_nameservers"]
                del self.data["kernel_options"]["netcfg/get_nameservers"]

    def _generate_static_ip_boot_options(self):
        """
        Try to add static ip boot options to avoid DHCP (interface/ip/netmask/gw/dns)
        Check for overrides first and clear them from kernel_options
        :return: The Tuple with the interface, IP, Netmask, Gateway and DNS information.
        """
        self._generate_static_ip_boot_interface()
        self._generate_static_ip_boot_ip()
        self._generate_static_ip_boot_mask()
        self._generate_static_ip_boot_gateway()
        self._generate_static_ip_boot_dns()

    def _generate_append_redhat(self):
        """
        Generate additional content for the append line in case that dist is a RedHat based one.
        """
        if self.data.get("proxy", "") != "":
            self.append_line += " proxy=%s http_proxy=%s" % (
                self.data["proxy"],
                self.data["proxy"],
            )
        self.append_line += " inst.ks=%s" % self.data["autoinstall"]

    def _generate_append_debian(self, system):
        """
        Generate additional content for the append line in case that dist is Ubuntu or Debian.
        :param system: The system which the append line should be generated for.
        """
        self.append_line += (
            " auto-install/enable=true url=%s netcfg/disable_autoconfig=true"
            % self.data["autoinstall"]
        )
        if self.data.get("proxy", "") != "":
            self.append_line += " mirror/http/proxy=%s" % self.data["proxy"]
        # hostname is required as a parameter, the one in the preseed is not respected
        my_domain = "local.lan"
        if system.hostname != "":
            # if this is a FQDN, grab the first bit
            my_hostname = system.hostname.split(".")[0]
            _domain = system.hostname.split(".")[1:]
            if _domain:
                my_domain = ".".join(_domain)
        else:
            my_hostname = system.name.split(".")[0]
            _domain = system.name.split(".")[1:]
            if _domain:
                my_domain = ".".join(_domain)
        # At least for debian deployments configured for DHCP networking this values are not used, but
        # specifying here avoids questions
        self.append_line += " hostname=%s domain=%s" % (my_hostname, my_domain)
        # A similar issue exists with suite name, as installer requires the existence of "stable" in the dists
        # directory
        self.append_line += " suite=%s" % self.dist.os_version

    def _generate_append_suse(self):
        """
        Special adjustments for generating the append line for suse.
        :return: The updated append line. If the distribution is not SUSE, then nothing is changed.
        """
        if self.data.get("proxy", "") != "":
            self.append_line += " proxy=%s" % self.data["proxy"]
        if self.data["kernel_options"].get("install", "") != "":
            self.append_line += " install=%s" % self.data["kernel_options"]["install"]
            del self.data["kernel_options"]["install"]
        else:
            self.append_line += " install=http://%s:%s/cblr/links/%s" % (
                self.data["server"],
                self.data["http_port"],
                self.dist.name,
            )
        if self.data["kernel_options"].get("autoyast", "") != "":
            self.append_line += " autoyast=%s" % self.data["kernel_options"]["autoyast"]
            del self.data["kernel_options"]["autoyast"]
        else:
            self.append_line += " autoyast=%s" % self.data["autoinstall"]

    def _adjust_interface_config(self):
        """
        If no kernel_options overrides are present find the management interface do nothing when zero or multiple
        management interfaces are found.
        """
        if self.system_interface is None:
            mgmt_ints = []
            mgmt_ints_multi = []
            slave_ints = []
            for iname, idata in self.data["interfaces"].items():
                if idata["management"] and idata["interface_type"] in [
                    "bond",
                    "bridge",
                ]:
                    # bonded/bridged management interface
                    mgmt_ints_multi.append(iname)
                if idata["management"] and idata["interface_type"] not in [
                    "bond",
                    "bridge",
                    "bond_slave",
                    "bridge_slave",
                    "bonded_bridge_slave",
                ]:
                    # single management interface
                    mgmt_ints.append(iname)

            if len(mgmt_ints_multi) == 1 and len(mgmt_ints) == 0:
                # Bonded/bridged management interface, find a slave interface if eth0 is a slave use that (it's what
                # people expect)
                for iname, idata in self.data["interfaces"].items():
                    if (
                        idata["interface_type"]
                        in ["bond_slave", "bridge_slave", "bonded_bridge_slave"]
                        and idata["interface_master"] == mgmt_ints_multi[0]
                    ):
                        slave_ints.append(iname)

                if "eth0" in slave_ints:
                    self.system_interface = "eth0"
                else:
                    self.system_interface = slave_ints[0]
                # Set self.system_ip from the bonded/bridged interface here
                self.system_ip = self.data[
                    "ip_address_"
                    + self.data["interface_master_" + self.system_interface]
                ]
                self.system_netmask = self.data[
                    "netmask_" + self.data["interface_master_" + self.system_interface]
                ]

            if len(mgmt_ints) == 1 and len(mgmt_ints_multi) == 0:
                # Single management interface
                self.system_interface = mgmt_ints[0]

    def _get_tcp_ip_config(self):
        """
        Lookup tcp/ip configuration data. If not present already present this adds it from the previously blenderd data.
        """
        if self.system_ip is None and self.system_interface is not None:
            intip = "ip_address_" + self.system_interface
            if self.data.get(intip, "") != "":
                self.system_ip = self.data["ip_address_" + self.system_interface]
        if self.system_netmask is None and self.system_interface is not None:
            intmask = "netmask_" + self.system_interface
            if self.data.get(intmask, "") != "":
                self.system_netmask = self.data["netmask_" + self.system_interface]
        if self.system_gw is None:
            if self.data.get("gateway", "") != "":
                self.system_gw = self.data["gateway"]
        if self.system_dns is None:
            if self.data.get("name_servers") != "":
                self.system_dns = self.data["name_servers"]

    def generate_system(self, dist, system, exclude_dns: bool) -> str:
        """
        Generate the append line for a netbooting system.
        :param dist: The distribution associated with the system.
        :param system: The system itself
        :param exclude_dns: Whether to include the DNS config or not.
        """
        self.dist = dist

        self.append_line = " append initrd=%s.img" % self.distro_name
        if self.dist.breed == "suse":
            self._generate_append_suse()
        elif self.dist.breed == "redhat":
            self._generate_append_redhat()
        elif dist.breed in ["ubuntu", "debian"]:
            self._generate_append_debian(system)

        self._generate_static_ip_boot_options()
        self._adjust_interface_config()
        self._get_tcp_ip_config()

        # Add information to the append_line
        self._system_int_append_line()
        self._system_ip_append_line()
        self._system_mask_append_line()
        self._system_gw_append_line()
        self._system_dns_append_line(exclude_dns)

        # Add remaining kernel_options to append_line
        self.append_line += buildiso.add_remaining_kopts(self.data["kernel_options"])
        return self.append_line

    def generate_profile(self, distro_breed: str) -> str:
        """
        Generate the append line for the kernel for a network installation.
        :param distro_breed: The name of the distribution breed.
        :return: The generated append line.
        """
        self.append_line = " append initrd=%s.img" % self.distro_name
        if distro_breed == "suse":
            if self.data.get("proxy", "") != "":
                self.append_line += " proxy=%s" % self.data["proxy"]
            if self.data["kernel_options"].get("install", "") != "":
                install_options = self.data["kernel_options"]["install"]
                if isinstance(install_options, list):
                    install_options = install_options[0]
                    self.append_line += " install=%s" % install_options
                del self.data["kernel_options"]["install"]
            else:
                self.append_line += " install=http://%s:%s/cblr/links/%s" % (
                    self.data["server"],
                    self.data["http_port"],
                    self.distro_name,
                )
            if self.data["kernel_options"].get("autoyast", "") != "":
                self.append_line += (
                    " autoyast=%s" % self.data["kernel_options"]["autoyast"]
                )
                del self.data["kernel_options"]["autoyast"]
            else:
                self.append_line += " autoyast=%s" % self.data["autoinstall"]
        elif distro_breed == "redhat":
            if self.data.get("proxy", "") != "":
                self.append_line += " proxy=%s http_proxy=%s" % (
                    self.data["proxy"],
                    self.data["proxy"],
                )
            self.append_line += " inst.ks=%s" % self.data["autoinstall"]
        elif distro_breed in ["ubuntu", "debian"]:
            self.append_line += (
                " auto-install/enable=true url=%s" % self.data["autoinstall"]
            )
            if self.data.get("proxy", "") != "":
                self.append_line += " mirror/http/proxy=%s" % self.data["proxy"]
        self.append_line += buildiso.add_remaining_kopts(self.data["kernel_options"])
        return self.append_line


class NetbootBuildiso(buildiso.BuildIso):
    """
    This class contains all functionality related to building network installation images.
    """

    def filter_systems(self, selected_items: List[str] = None) -> list:
        """
        Return a list of valid system objects selected from all systems by name, or everything if ``selected_items`` is
        empty.
        :param selected_items: A list of names to include in the returned list.
        :return: A list of valid systems. If an error occurred this is logged and an empty list is returned.
        """
        if selected_items is None:
            selected_items = []
        return self.filter_items(self.api.systems(), selected_items)

    def make_shorter(self, distname: str) -> str:
        """
        Return a short distro identifier which is basically an internal counter which is mapped via the real distro
        name.
        :param distname: The distro name to return an identifier for.
        :return: A short distro identifier
        """
        if distname in self.distmap:
            return self.distmap[distname]

        self.distctr += 1
        self.distmap[distname] = str(self.distctr)
        return str(self.distctr)

    def _generate_netboot_system(self, system, cfglines: List[str], exclude_dns: bool):
        """
        Generates the ISOLINUX cfg configuration for any systems included in the image.
        :param system: The system which the configuration should be generated for.
        :param cfglines: The already existing lines of the configuration.
        :param exclude_dns: If DNS configuration should be excluded or not.
        """
        self.logger.info("processing system: %s", system.name)
        profile = system.get_conceptual_parent()
        dist = profile.get_conceptual_parent()
        distname = self.make_shorter(dist.name)
        self.copy_boot_files(dist, self.isolinuxdir, distname)

        cfglines.append("")
        cfglines.append("LABEL %s" % system.name)
        cfglines.append("  MENU LABEL %s" % system.name)
        cfglines.append("  kernel %s.krn" % distname)

        data = utils.blender(self.api, False, system)
        if not re.match(r"[a-z]+://.*", data["autoinstall"]):
            data["autoinstall"] = "http://%s:%s/cblr/svc/op/autoinstall/system/%s" % (
                data["server"],
                data["http_port"],
                system.name,
            )

        append_builder = AppendLineBuilder(distro_name=distname, data=data)
        append_line = append_builder.generate_system(dist, system, exclude_dns)
        cfglines.append(append_line)

    def _generate_netboot_profile(self, profile, cfglines: List[str]):
        """
        Generates the ISOLINUX cfg configuration for any profiles included in the image.
        :param profile: The profile which the configuration should be generated for.
        :param cfglines: The already existing lines of the configuration.
        """
        self.logger.info('Processing profile: "%s"', profile.name)
        dist = profile.get_conceptual_parent()
        distname = self.make_shorter(dist.name)
        self.copy_boot_files(dist, self.isolinuxdir, distname)

        cfglines.append("")
        cfglines.append("LABEL %s" % profile.name)
        cfglines.append("  MENU LABEL %s" % profile.name)
        cfglines.append("  kernel %s.krn" % distname)

        data = utils.blender(self.api, False, profile)

        # SUSE is not using 'text'. Instead 'textmode' is used as kernel option.
        if dist is not None:
            utils.kopts_overwrite(
                data["kernel_options"], self.api.settings().server, dist.breed
            )

        if not re.match(r"[a-z]+://.*", data["autoinstall"]):
            data["autoinstall"] = "http://%s:%s/cblr/svc/op/autoinstall/profile/%s" % (
                data["server"],
                data["http_port"],
                profile.name,
            )

        append_builder = AppendLineBuilder(distro_name=distname, data=data)
        append_line = append_builder.generate_profile(dist.breed)
        cfglines.append(append_line)

    def generate_netboot_iso(
        self, systems: List[str] = None, exclude_dns: bool = False
    ):
        """
        Creates the ``isolinux.cfg`` for a network bootable ISO image.
        :param systems: The filter to generate a netboot iso for. You may specify multiple systems on the CLI space
                        separated.
        :param exclude_dns: If this is True then the dns server is skipped. False will set it.
        """
        # setup isolinux.cfg
        isolinuxcfg = os.path.join(self.isolinuxdir, "isolinux.cfg")
        cfglines = [self.iso_template]

        # iterate through selected profiles
        for profile in self.filter_profiles(self.profiles):
            self._generate_netboot_profile(profile, cfglines)
        cfglines.append("MENU SEPARATOR")
        # iterate through all selected systems
        for system in self.filter_systems(systems):
            self._generate_netboot_system(system, cfglines, exclude_dns)
        cfglines.append("")
        cfglines.append("MENU END")

        with open(isolinuxcfg, "w+") as cfg:
            cfg.writelines(cfglines)

    def run(
        self,
        iso: str = "autoinst.iso",
        buildisodir: str = "",
        profiles: List[str] = None,
        xorrisofs_opts: str = "",
        distro_name: str = "",
        systems: List[str] = None,
        exclude_dns: bool = False,
    ):
        """
        Run the whole iso generation from bottom to top. Per default this builds an ISO for all available systems
        and profiles.
        This is the only method which should be called from non-class members. The ``profiles`` and ``system``
        parameters can be combined.
        :param iso: The name of the iso. Defaults to "autoinst.iso".
        :param buildisodir: This overwrites the directory from the settings in which the iso is built in.
        :param profiles: The filter to generate the ISO only for selected profiles.
        :param xorrisofs_opts: ``xorrisofs`` options to include additionally.
        :param distro_name: For detecting the architecture of the ISO.
        :param systems: Don't use that when building standalone ISOs. The filter to generate the ISO only for selected
                        systems.
        :param exclude_dns: Whether the repositories have to be locally available or the internet is reachable.
        """
        buildisodir = self._prepare_iso(buildisodir, distro_name, profiles)
        systems = utils.input_string_or_list_no_inherit(systems)
        self.generate_netboot_iso(systems, exclude_dns)
        self._generate_iso(xorrisofs_opts, iso, buildisodir)



from xmlrpc.client import Fault

import pytest


@pytest.mark.usefixtures("cobbler_xmlrpc_base")
class TestSystem:
    """
    Test remote calls related to systems.
    """

    def test_get_systems(self, remote, token):
        """
        Test: get systems
        """
        # Arrange --> Nothing to arrange

        # Act
        result = remote.get_systems(token)

        # Assert
        assert result == []

    @pytest.mark.usefixtures(
        "create_testdistro",
        "create_testmenu",
        "create_testprofile",
        "remove_testdistro",
        "remove_testmenu",
        "remove_testprofile",
        "remove_testsystem",
    )
    @pytest.mark.parametrize(
        "field_name,field_value",
        [
            ("comment", "test comment"),
            ("enable_ipxe", True),
            ("enable_ipxe", False),
            ("kernel_options", "a=1 b=2 c=3 c=4 c=5 d e"),
            ("kernel_options_post", "a=1 b=2 c=3 c=4 c=5 d e"),
            ("autoinstall", "test.ks"),
            ("autoinstall", "test.xml"),
            ("autoinstall", "test.seed"),
            ("autoinstall_meta", "a=1 b=2 c=3 c=4 c=5 d e"),
            ("mgmt_classes", "one two three"),
            ("mgmt_parameters", "<<inherit>>"),
            ("name", "testsystem0"),
            ("netboot_enabled", True),
            ("netboot_enabled", False),
            ("owners", "user1 user2 user3"),
            ("profile", "testprofile0"),
            ("repos_enabled", True),
            ("repos_enabled", False),
            ("status", "development"),
            ("status", "testing"),
            ("status", "acceptance"),
            ("status", "production"),
            ("proxy", "testproxy"),
            ("server", "1.1.1.1"),
            # ("boot_loaders", "pxe ipxe grub"), FIXME: This raises currently but it did not in the past
            ("virt_auto_boot", True),
            ("virt_auto_boot", False),
            ("virt_auto_boot", "yes"),
            ("virt_auto_boot", "no"),
            ("virt_cpus", "<<inherit>>"),
            ("virt_cpus", 1),
            ("virt_cpus", 2),
            ("virt_cpus", "<<inherit>>"),
            ("virt_file_size", "<<inherit>>"),
            ("virt_file_size", 5),
            ("virt_file_size", 10),
            ("virt_disk_driver", "<<inherit>>"),
            ("virt_disk_driver", "raw"),
            ("virt_disk_driver", "qcow2"),
            ("virt_disk_driver", "vdmk"),
            ("virt_ram", "<<inherit>>"),
            ("virt_ram", 256),
            ("virt_ram", 1024),
            ("virt_type", "<<inherit>>"),
            ("virt_type", "xenpv"),
            ("virt_type", "xenfv"),
            ("virt_type", "qemu"),
            ("virt_type", "kvm"),
            ("virt_type", "vmware"),
            ("virt_type", "openvz"),
            ("virt_path", "<<inherit>>"),
            ("virt_path", "/path/to/test"),
            ("virt_pxe_boot", True),
            ("virt_pxe_boot", False),
            ("power_type", "ipmilanplus"),
            ("power_address", "127.0.0.1"),
            ("power_id", "pmachine:lpar1"),
            ("power_pass", "pass"),
            ("power_user", "user"),
        ],
    )
    def test_create_system_positive(
        self, remote, token, template_files, field_name, field_value
    ):
        """
        Test: create/edit a system object
        """
        # Arrange
        system = remote.new_system(token)
        remote.modify_system(system, "name", "testsystem0", token)
        remote.modify_system(system, "profile", "testprofile0", token)

        # Act
        result = remote.modify_system(system, field_name, field_value, token)

        # Assert
        assert result

    @pytest.mark.usefixtures(
        "create_testdistro",
        "create_testmenu",
        "create_testprofile",
        "remove_testdistro",
        "remove_testmenu",
        "remove_testprofile",
        "remove_testsystem",
    )
    @pytest.mark.parametrize(
        "field_name,field_value",
        [
            ("autoinstall", "/path/to/bad/autoinstall"),
            ("mgmt_parameters", "badyaml"),
            ("profile", "badprofile"),
            ("boot_loaders", "badloader"),
            ("virt_cpus", "a"),
            ("virt_file_size", "a"),
            ("virt_ram", "a"),
            ("virt_type", "bad"),
            ("power_type", "bla"),
        ],
    )
    def test_create_system_negative(self, remote, token, field_name, field_value):
        """
        Test: create/edit a system object
        """
        # Arrange
        system = remote.new_system(token)
        remote.modify_system(system, "name", "testsystem0", token)
        remote.modify_system(system, "profile", "testprofile0", token)

        # Act & Assert
        try:
            remote.modify_system(system, field_name, field_value, token)
        except Fault:
            assert True
        else:
            pytest.fail("Bad field did not raise an exception!")

    def test_find_system(self, remote, token):
        """
        Test: find a system object
        """

        # Arrange --> Nothing to arrange

        # Act
        result = remote.find_system({"name": "notexisting"}, token)

        # Assert --> A not exiting system returns an empty list
        assert result == []

    @pytest.mark.usefixtures(
        "create_testdistro",
        "create_testmenu",
        "create_testprofile",
        "remove_testdistro",
        "remove_testmenu",
        "remove_testprofile",
        "remove_testsystem",
    )
    def test_add_interface_to_system(self, remote, token):
        """
        Test: add an interface to a system
        """

        # Arrange
        system = remote.new_system(token)
        remote.modify_system(system, "name", "testsystem0", token)
        remote.modify_system(system, "profile", "testprofile0", token)

        # Act
        result = remote.modify_system(
            system, "modify_interface", {"macaddress-eth0": "aa:bb:cc:dd:ee:ff"}, token
        )
        remote.save_system(system, token)

        # Assert --> returns true if successful
        assert result
        assert (
            remote.get_system("testsystem0")
            .get("interfaces", {})
            .get("eth0", {})
            .get("mac_address")
            == "aa:bb:cc:dd:ee:ff"
        )

    @pytest.mark.usefixtures(
        "create_testdistro",
        "create_testmenu",
        "create_testprofile",
        "remove_testdistro",
        "remove_testmenu",
        "remove_testprofile",
        "remove_testsystem",
    )
    def test_remove_interface_from_system(self, remote, token):
        """
        Test: remove an interface from a system
        """

        # Arrange
        system = remote.new_system(token)
        remote.modify_system(system, "name", "testsystem0", token)
        remote.modify_system(system, "profile", "testprofile0", token)
        remote.modify_system(
            system, "modify_interface", {"macaddress-eth0": "aa:bb:cc:dd:ee:ff"}, token
        )
        remote.save_system(system, token)

        # Act
        result = remote.modify_system(
            system, "delete_interface", {"interface": "eth0"}, token
        )
        remote.save_system(system, token)

        # Assert --> returns true if successful
        assert result
        assert (
            remote.get_system("testsystem0").get("interfaces", {}).get("eth0", None)
            is None
        )

    @pytest.mark.usefixtures(
        "create_testdistro",
        "create_testmenu",
        "create_testprofile",
        "remove_testdistro",
        "remove_testmenu",
        "remove_testprofile",
        "remove_testsystem",
    )
    def test_rename_interface(self, remote, token):
        """
        Test: rename an interface on a system
        """

        # Arrange
        system = remote.new_system(token)
        remote.modify_system(system, "name", "testsystem0", token)
        remote.modify_system(system, "profile", "testprofile0", token)
        result_add = remote.modify_system(
            system, "modify_interface", {"macaddress-eth0": "aa:bb:cc:dd:ee:ff"}, token
        )
        remote.save_system(system, token)

        # Act
        result_rename = remote.modify_system(
            system,
            "rename_interface",
            {"interface": "eth0", "rename_interface": "eth_new"},
            token,
        )
        remote.save_system(system, token)

        # Assert --> returns true if successful
        assert result_add
        assert result_rename
        assert (
            remote.get_system("testsystem0").get("interfaces", {}).get("eth0", None)
            is None
        )
        assert (
            remote.get_system("testsystem0")
            .get("interfaces", {})
            .get("eth_new", {})
            .get("mac_address")
            == "aa:bb:cc:dd:ee:ff"
        )

    @pytest.mark.usefixtures(
        "create_testdistro",
        "create_testmenu",
        "create_testprofile",
        "create_testsystem",
        "remove_testdistro",
        "remove_testmenu",
        "remove_testprofile",
        "remove_testsystem",
    )
    def test_copy_system(self, remote, token):
        """
        Test: copy a system object
        """

        # Arrange
        system = remote.get_item_handle("system", "testsystem0", token)

        # Act
        result = remote.copy_system(system, "testsystemcopy", token)

        # Assert
        assert result

        # Cleanup
        remote.remove_system("testsytemcopy", token)

    @pytest.mark.usefixtures(
        "create_testdistro",
        "create_testmenu",
        "create_testprofile",
        "create_testsystem",
        "remove_testdistro",
        "remove_testmenu",
        "remove_testprofile",
        "remove_testsystem",
    )
    def test_rename_system(self, remote, token):
        """
        Test: rename a system object
        """

        # Arrange --> Done in fixtures also.
        system = remote.get_item_handle("system", "testsystem0", token)

        # Act
        result = remote.rename_system(system, "testsystem1", token)

        # Assert
        assert result

        # Cleanup
        remote.remove_system("testsystem1", token)

    def test_remove_system(self, remote, token):
        """
        Test: remove a system object
        """

        # Arrange
        # TODO: Verify why the test passes without the fixture for creating the system!

        # Act
        result = remote.remove_system("testsystem0", token)

        # Assert
        assert result

    def test_get_repo_config_for_system(self, remote):
        """
        Test: get repository configuration of a system
        """

        # Arrange --> There is nothing to be arranged

        # Act
        result = remote.get_repo_config_for_system("testprofile0")

        # Assert --> Let the test pass if the call is okay.
        assert True



import pytest
from cobbler.api import CobblerAPI
from cobbler.settings import Settings
from cobbler.modules.authentication import ldap

#@pytest.fixture(scope="class")
@pytest.fixture()
def api():
    return CobblerAPI()

@pytest.fixture()
def test_settings(api):
    settings = api.settings()
    settings.ldap_server = "localhost"
    settings.ldap_port = 389
    settings.ldap_base_dn = "dc=example,dc=com"
    settings.ldap_search_prefix = "uid="
    settings.ldap_anonymous_bind = True
    settings.ldap_reqcert = "hard"
    return settings

class TestLdap:
    @pytest.mark.parametrize("anonymous_bind, username, password", [
        (True, "test", "test")
    ])
    def test_anon_bind_positive(self, api, test_settings, anonymous_bind, username, password):
        # Arrange
        test_settings.ldap_anonymous_bind = anonymous_bind
        test_settings.ldap_tls = False

        # Act
        result = ldap.authenticate(api, username, password)

        # Assert
        assert result

    @pytest.mark.parametrize("anonymous_bind, username, password", [
        (True, "test", "bad")
    ])
    def test_anon_bind_negative(self, api, test_settings, anonymous_bind, username, password):
        # Arrange
        test_settings.ldap_anonymous_bind = anonymous_bind
        test_settings.ldap_tls = False

        # Act
        result = ldap.authenticate(api, username, password)

        # Assert
        assert not result

    @pytest.mark.parametrize("anonymous_bind, bind_user, bind_password, username, password", [
        (False, "uid=user,dc=example,dc=com", "test", "test", "test")
    ])
    def test_user_bind_positive(self, api, test_settings, anonymous_bind, bind_user, bind_password, username, password):
        # Arrange
        test_settings.ldap_anonymous_bind = anonymous_bind
        test_settings.ldap_search_bind_dn = bind_user
        test_settings.ldap_search_passwd = bind_password
        test_settings.ldap_tls = False

        # Act
        result = ldap.authenticate(api, username, password)

        # Assert
        assert result

    @pytest.mark.parametrize("anonymous_bind, bind_user, bind_password, username, password", [
        (False, "uid=user,dc=example,dc=com", "bad", "test", "test")
    ])
    def test_user_bind_negative(self, api, test_settings, anonymous_bind, bind_user, bind_password, username, password):
        # Arrange
        test_settings.ldap_anonymous_bind = anonymous_bind
        test_settings.ldap_search_bind_dn = bind_user
        test_settings.ldap_search_passwd = bind_password
        test_settings.ldap_tls = False

        # Act
        result = ldap.authenticate(api, username, password)

        # Assert
        assert not result

    @pytest.mark.parametrize("tls_cadir, tls_cert, tls_key", [
        ("/etc/ssl/certs",
         "/etc/ssl/ldap.crt",
         "/etc/ssl/ldap.key")
    ])
    def test_cadir_positive(self, api, test_settings, tls_cadir, tls_cert, tls_key):
        # Arrange
        test_settings.ldap_tls = True
        test_settings.ldap_tls_cacertdir = tls_cadir
        test_settings.ldap_tls_cacertfile = None
        test_settings.ldap_tls_certfile = tls_cert
        test_settings.ldap_tls_keyfile = tls_key

        # Act
        result = ldap.authenticate(api, "test", "test")

        # Assert
        assert result

    @pytest.mark.parametrize("tls_cadir, tls_cert, tls_key", [
        ("/etc/ssl/certs",
         "/etc/ssl/bad.crt",
         "/etc/ssl/bad.key")
    ])
    def test_cadir_negative(self, api, test_settings, tls_cadir, tls_cert, tls_key):
        # Arrange
        test_settings.ldap_tls = True
        test_settings.ldap_tls_cacertdir = tls_cadir
        test_settings.ldap_tls_cacertfile = None
        test_settings.ldap_tls_certfile = tls_cert
        test_settings.ldap_tls_keyfile = tls_key

        # Act
        result = ldap.authenticate(api, "test", "test")

        # Assert
        assert not result

    @pytest.mark.parametrize("tls_cafile, tls_cert, tls_key", [
        ("/etc/ssl/ca-slapd.crt",
         "/etc/ssl/ldap.crt",
         "/etc/ssl/ldap.key")
    ])
    def test_cafile_positive(self, api, test_settings, tls_cafile, tls_cert, tls_key):
        # Arrange
        test_settings.ldap_tls = True
        test_settings.ldap_tls_cacertdir = None
        test_settings.ldap_tls_cacertfile = tls_cafile
        test_settings.ldap_tls_certfile = tls_cert
        test_settings.ldap_tls_keyfile = tls_key

        # Act
        result = ldap.authenticate(api, "test", "test")

        # Assert
        assert result

    @pytest.mark.parametrize("tls_cafile, tls_cert, tls_key", [
        ("/etc/ssl/ca-slapd.crt",
         "/etc/ssl/bad.crt",
         "/etc/ssl/bad.key")
    ])
    def test_cafile_negative(self, api, test_settings, tls_cafile, tls_cert, tls_key):
        # Arrange
        test_settings.ldap_tls = True
        test_settings.ldap_tls_cacertdir = None
        test_settings.ldap_tls_cacertfile = tls_cafile
        test_settings.ldap_tls_certfile = tls_cert
        test_settings.ldap_tls_keyfile = tls_key

        # Act
        result = ldap.authenticate(api, "test", "test")

        # Assert
        assert not result

    @pytest.mark.parametrize("tls_cafile, tls_cert, tls_key", [
        ("/etc/ssl/ca-slapd.crt",
         "/etc/ssl/ldap.crt",
         "/etc/ssl/ldap.key")
    ])
    def test_ldaps_positive(self, api, test_settings, tls_cafile, tls_cert, tls_key):
        # Arrange
        test_settings.ldap_tls = False
        test_settings.ldap_port = 636
        test_settings.ldap_tls_cacertdir = None
        test_settings.ldap_tls_cacertfile = tls_cafile
        test_settings.ldap_tls_certfile = tls_cert
        test_settings.ldap_tls_keyfile = tls_key

        # Act
        result = ldap.authenticate(api, "test", "test")

        # Assert
        assert result

    @pytest.mark.parametrize("tls_cafile, tls_cert, tls_key", [
        ("/etc/ssl/ca-slapd.crt",
         "/etc/ssl/bad.crt",
         "/etc/ssl/bad.key")
    ])
    def test_ldaps_negative(self, api, test_settings, tls_cafile, tls_cert, tls_key):
        # Arrange
        test_settings.ldap_tls = False
        test_settings.ldap_port = 636
        test_settings.ldap_tls_cacertdir = None
        test_settings.ldap_tls_cacertfile = tls_cafile
        test_settings.ldap_tls_certfile = tls_cert
        test_settings.ldap_tls_keyfile = tls_key

        # Act & Assert
        with pytest.raises(ValueError):
            result = ldap.authenticate(api, "test", "test")



"""Cobbler action to create bootable Grub2 images.

This action calls grub2-mkimage for all bootloader formats configured in
Cobbler's settings. See man(1) grub2-mkimage for available formats.
"""
import logging
import pathlib
import re
import subprocess
import sys
import typing

from cobbler import utils


# NOTE: does not warrant being a class, but all Cobbler actions use a class's ".run()" as the entrypoint
class MkLoaders:
    """
    Action to create bootloader images.
    """

    def __init__(self, api):
        """
        MkLoaders constructor.

        :param api: CobblerAPI instance for accessing settings
        """
        self.logger = logging.getLogger()
        self.bootloaders_dir = pathlib.Path(api.settings().bootloaders_dir)
        # GRUB 2
        self.grub2_mod_dir = pathlib.Path(api.settings().grub2_mod_dir)
        self.boot_loaders_formats: typing.Dict = api.settings().bootloaders_formats
        self.modules: typing.List = api.settings().bootloaders_modules
        # Syslinux
        self.syslinux_folder = pathlib.Path(api.settings().syslinux_dir)
        self.syslinux_memdisk_folder = pathlib.Path(api.settings().syslinux_memdisk_folder)
        self.syslinux_pxelinux_folder = pathlib.Path(api.settings().syslinux_pxelinux_folder)
        # Shim
        self.shim_glob = pathlib.Path(api.settings().bootloaders_shim_folder)
        self.shim_regex = re.compile(api.settings().bootloaders_shim_file)
        # iPXE
        self.ipxe_folder = pathlib.Path(api.settings().bootloaders_ipxe_folder)

    def run(self):
        """
        Run GrubImages action. If the files or executables for the bootloader is not available we bail out and skip the
        creation after it is logged that this is not available.
        """
        self.create_directories()

        self.make_shim()
        self.make_ipxe()
        self.make_syslinux()
        self.make_grub()

    def make_shim(self):
        """
        Create symlink of the shim bootloader in case it is available on the system.
        """
        # Check well-known locations
        # Absolute paths are not supported BUT we can get around that: https://stackoverflow.com/a/51108375/4730773
        parts = self.shim_glob.parts
        start_at = 1 if self.shim_glob.is_absolute() else 0
        bootloader_path_parts = pathlib.Path(*parts[start_at:])
        results = sorted(pathlib.Path(self.shim_glob.root).glob(str(bootloader_path_parts)))
        # If no match, then report and bail out.
        if len(results) <= 0:
            self.logger.info('Unable to find the folder which should be scanned for "shim.efi"! Bailing out of linking '
                             'the shim!')
            return
        # Now scan the folders with the regex
        target_shim = None
        for possible_folder in results:
            for child in possible_folder.iterdir():
                if self.shim_regex.search(str(child)):
                    target_shim = child.resolve()
                    break
        # If no match is found report and return
        if target_shim is None:
            self.logger.info('Unable to find "shim.efi" file. Please adjust "bootloaders_shim_file" regex. Bailing out '
                             'of linking the shim!')
            return
        # Symlink the absolute target of the match
        symlink(
            target_shim,
            self.bootloaders_dir.joinpath(pathlib.Path("grub/shim.efi")),
            skip_existing=True
        )

    def make_ipxe(self):
        """
        Create symlink of the iPXE bootloader in case it is available on the system.
        """
        if not self.ipxe_folder.exists():
            self.logger.info('ipxe directory did not exist. Please adjust the "bootloaders_ipxe_folder". Bailing out '
                             'of iPXE setup!')
            return
        symlink(
            self.ipxe_folder.joinpath("undionly.kpxe"),
            self.bootloaders_dir.joinpath(pathlib.Path("undionly.pxe")),
            skip_existing=True
        )

    def make_syslinux(self):
        """
        Create symlink of the important syslinux bootloader files in case they are available on the system.
        """
        if not utils.command_existing("syslinux"):
            self.logger.info("syslinux command not available. Bailing out of syslinux setup!")
            return
        # Make modules
        symlink(
            self.syslinux_folder.joinpath("menu.c32"),
            self.bootloaders_dir.joinpath("menu.c32"),
            skip_existing=True
        )
        if get_syslinux_version() < 5:
            # This file is only required for Syslinux 5 and newer.
            # Source: https://wiki.syslinux.org/wiki/index.php?title=Library_modules
            self.logger.info('syslinux version 4 detected! Skip making symlink of "ldlinux.c32" file!')
        else:
            symlink(
                self.syslinux_folder.joinpath("ldlinux.c32"),
                self.bootloaders_dir.joinpath("ldlinux.c32"),
                skip_existing=True
            )
        # Make memdisk
        symlink(
            self.syslinux_memdisk_folder.joinpath("memdisk"),
            self.bootloaders_dir.joinpath("memdisk"),
            skip_existing=True
        )
        # Make pxelinux.0
        symlink(
            self.syslinux_pxelinux_folder.joinpath("pxelinux.0"),
            self.bootloaders_dir.joinpath("pxelinux.0"),
            skip_existing=True
        )

    def make_grub(self):
        """
        Create symlink of the GRUB 2 bootloader in case it is available on the system. Additionally build the loaders
        for other architectures if the modules to do so are available.
        """
        if not utils.command_existing("grub2-mkimage"):
            self.logger.info("grub2-mkimage command not available. Bailing out of GRUB2 generation!")
            return

        for image_format, options in self.boot_loaders_formats.items():
            bl_mod_dir = options.get("mod_dir", image_format)
            mod_dir = self.grub2_mod_dir.joinpath(bl_mod_dir)
            if not mod_dir.exists():
                self.logger.info(
                    'GRUB2 modules directory for arch "%s" did no exist. Skipping GRUB2 creation',
                    image_format
                )
                continue
            try:
                mkimage(
                    image_format,
                    self.bootloaders_dir.joinpath("grub", options["binary_name"]),
                    self.modules + options.get("extra_modules", []),
                )
            except subprocess.CalledProcessError:
                self.logger.info('grub2-mkimage failed for arch "%s"! Maybe you did forget to install the grub modules '
                                 'for the architecture?', image_format)
                utils.log_exc()
                # don't create module symlinks if grub2-mkimage is unsuccessful
                continue
            self.logger.info('Successfully built bootloader for arch "%s"!', image_format)

            # Create a symlink for GRUB 2 modules
            # assumes a single GRUB can be used to boot all kinds of distros
            # if this assumption turns out incorrect, individual "grub" subdirectories are needed
            symlink(
                mod_dir,
                self.bootloaders_dir.joinpath("grub", bl_mod_dir),
                skip_existing=True
            )

    def create_directories(self):
        """
        Create the required directories so that this succeeds. If existing, do nothing. This should create the tree for
        all supported bootloaders, regardless of the capabilities to symlink/install/build them.
        """
        if not self.bootloaders_dir.exists():
            raise FileNotFoundError("Main bootloader directory not found! Please create it yourself!")

        grub_dir = self.bootloaders_dir.joinpath("grub")
        if not grub_dir.exists():
            grub_dir.mkdir(mode=0o644)


# NOTE: move this to cobbler.utils?
# cobbler.utils.linkfile does a lot of things, it might be worth it to have a
# function just for symbolic links
def symlink(target: pathlib.Path, link: pathlib.Path, skip_existing: bool = False):
    """Create a symlink LINK pointing to TARGET.

    :param target: File/directory that the link will point to. The file/directory must exist.
    :param link: Filename for the link.
    :param skip_existing: Controls if existing links are skipped, defaults to False.
    :raises FileNotFoundError: ``target`` is not an existing file.
    :raises FileExistsError: ``skip_existing`` is False and ``link`` already exists.
    """

    if not target.exists():
        raise FileNotFoundError(
            f"{target} does not exist, can't create a symlink to it."
        )
    try:
        link.symlink_to(target)
    except FileExistsError:
        if not skip_existing:
            raise


def mkimage(image_format: str, image_filename: pathlib.Path, modules: typing.List):
    """Create a bootable image of GRUB using grub2-mkimage.

    :param image_format: Format of the image that is being created. See man(1)
        grub2-mkimage for a list of supported formats.
    :param image_filename: Location of the image that is being created.
    :param modules: List of GRUB modules to include into the image
    :raises subprocess.CalledProcessError: Error raised by ``subprocess.run``.
    """

    if not image_filename.parent.exists():
        image_filename.parent.mkdir(parents=True)

    cmd = ["grub2-mkimage"]
    cmd.extend(("--format", image_format))
    cmd.extend(("--output", str(image_filename)))
    cmd.append("--prefix=")
    cmd.extend(modules)

    # The Exception raised by subprocess already contains everything useful, it's simpler to use that than roll our
    # own custom exception together with cobbler.utils.subprocess_* functions
    subprocess.run(cmd, check=True)


def get_syslinux_version() -> int:
    """
    This calls syslinux and asks for the version number.

    :return: The major syslinux release number.
    :raises subprocess.CalledProcessError: Error raised by ``subprocess.run`` in case syslinux does not return zero.
    """
    # Example output: "syslinux 4.04  Copyright 1994-2011 H. Peter Anvin et al"
    cmd = ["syslinux", "-v"]
    completed_process = subprocess.run(cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                       encoding=sys.getdefaultencoding())
    output = completed_process.stdout.split()
    return int(float(output[1]))



import pytest

from cobbler import enums
from tests.conftest import does_not_raise


@pytest.mark.parametrize(
    "test_architecture,test_raise",
    [
        (enums.Archs.X86_64, does_not_raise()),
        ("x86_64", does_not_raise()),
        ("abc", pytest.raises(ValueError)),
        (0, pytest.raises(TypeError)),
    ],
)
def test_validate_arch(test_architecture, test_raise):
    # Arrange

    # Act
    with test_raise:
        result = enums.Archs.to_enum(test_architecture)

        # Assert
        if isinstance(test_architecture, str):
            assert result.value == test_architecture
        elif isinstance(test_architecture, enums.Archs):
            assert result == test_architecture
        else:
            raise TypeError("result had a non expected result")


@pytest.mark.parametrize("value,expected_exception", [
    ("qemu", does_not_raise()),
    ("<<inherit>>", does_not_raise()),
    (enums.VirtType.QEMU, does_not_raise()),
    (enums.VirtType.INHERITED, does_not_raise()),
    (0, pytest.raises(TypeError))
])
def test_set_virt_type(value, expected_exception):
    # Arrange

    # Act
    with expected_exception:
        result = enums.VirtType.to_enum(value)

        # Assert
        if isinstance(value, str):
            assert result.value == value
        elif isinstance(value, enums.VirtType):
            assert result == value
        else:
            raise TypeError("Unexpected type for value!")


@pytest.mark.parametrize("value,expected_exception", [
    ("allow", does_not_raise()),
    (enums.TlsRequireCert.ALLOW, does_not_raise()),
    (enums.VALUE_INHERITED, pytest.raises(ValueError)),
    (0, pytest.raises(TypeError))
])
def test_validate_ldap_tls_reqcert(value, expected_exception):
    # Arrange

    # Act
    with expected_exception:
        result = enums.TlsRequireCert.to_enum(value)

        # Assert
        if isinstance(value, str):
            assert result.value == value
        elif isinstance(value, enums.TlsRequireCert):
            assert result == value
        else:
            raise TypeError("Unexpected type for value!")


@pytest.mark.parametrize("test_driver,test_raise", [
    (enums.VirtDiskDrivers.RAW, does_not_raise()),
    (enums.VALUE_INHERITED, does_not_raise()),
    (enums.VirtDiskDrivers.INHERITED, does_not_raise()),
    ("qcow2", does_not_raise()),
    ("<<inherit>>", does_not_raise()),
    ("bad_driver", pytest.raises(ValueError)),
    (0, pytest.raises(TypeError))
])
def test_set_virt_disk_driver(test_driver, test_raise):
    # Arrange

    # Act
    with test_raise:
        result = enums.VirtDiskDrivers.to_enum(test_driver)

        # Assert
        if isinstance(test_driver, str):
            assert result.value == test_driver
        elif isinstance(test_driver, enums.VirtDiskDrivers):
            assert result == test_driver
        else:
            raise TypeError("Unexpected type for value!")



"""
Builds bootable CD images that have PXE-equivalent behavior for all Cobbler distros/profiles/systems currently in
memory.
"""

# SPDX-License-Identifier: GPL-2.0-or-later
# SPDX-FileCopyrightText: Copyright 2006-2009, Red Hat, Inc and Others
# SPDX-FileCopyrightText: Michael DeHaan <michael.dehaan AT gmail>

import logging
import os
import pathlib
import shutil
from typing import Dict, List, Optional, Union

from cobbler import utils
from cobbler.enums import Archs


def add_remaining_kopts(kopts: dict) -> str:
    """Add remaining kernel_options to append_line
    :param kopts: The kernel options which are not present in append_line.
    :return: A single line with all kernel options from the dictionary in the string. Starts with a space.
    """
    append_line = [""]  # empty str to ensure the returned str starts with a space
    for option, args in kopts.items():
        if args is None:
            append_line.append(f"{option}")
            continue

        if not isinstance(args, list):
            args = [args]

        for arg in args:
            arg_str = format(arg)
            if " " in arg_str:
                arg_str = f'"{arg_str}"'
            append_line.append(f"{option}={arg_str}")

    return " ".join(append_line)


class BuildIso:
    """
    Handles conversion of internal state to the isolinux tree layout
    """

    def __init__(self, api):
        """Constructor which initializes things here. The collection manager pulls all other dependencies in.
        :param api: The API instance which holds all information about objects in Cobbler.
        """
        self.api = api
        self.distmap = {}
        self.distctr = 0
        self.logger = logging.getLogger()

        self.profiles: List[str] = []
        self.isolinuxdir = ""

        # grab the header from buildiso.header file
        self.iso_template = (
            pathlib.Path(self.api.settings().iso_template_dir)
            .joinpath("buildiso.template")
            .read_text()
        )

    def copy_boot_files(self, distro, destdir: str, new_filename: str = ""):
        """Copy kernel/initrd to destdir with (optional) newfile prefix
        :param distro: Distro object to return the boot files for.
        :param destdir: The destination directory.
        :param new_filename: The file new filename. Kernel and Initrd have different extensions to seperate them from
                             each another.
        """
        self.logger.info("copying kernels and initrds for standalone distro")
        kernel = pathlib.Path(distro.kernel)
        initrd = pathlib.Path(distro.initrd)
        path_destdir = pathlib.Path(destdir)
        if not new_filename:
            shutil.copyfile(kernel, path_destdir.joinpath(kernel.name))
            shutil.copyfile(initrd, path_destdir.joinpath(initrd.name))
        else:
            shutil.copyfile(kernel, path_destdir.joinpath("%s.krn" % new_filename))
            shutil.copyfile(initrd, path_destdir.joinpath("%s.img" % new_filename))

    def filter_profiles(self, selected_items: List[str] = None) -> list:
        """
        Return a list of valid profile objects selected from all profiles by name, or everything if ``selected_items``
        is empty.
        :param selected_items: A list of names to include in the returned list.
        :return: A list of valid profiles. If an error occurred this is logged and an empty list is returned.
        """
        if selected_items is None:
            selected_items = []
        return self.filter_items(self.api.profiles(), selected_items)

    def filter_items(self, all_objs, selected_items: List[str]) -> list:
        """Return a list of valid profile or system objects selected from all profiles or systems by name, or everything
        if selected_items is empty.
        :param all_objs: The collection of items to filter.
        :param selected_items: The list of names
        :raises ValueError: Second option that this error is raised
                            when the list of filtered systems or profiles is empty.
        :return: A list of valid profiles OR systems. If an error occurred this is logged and an empty list is returned.
        """
        # No profiles/systems selection is made, let's return everything.
        if len(selected_items) == 0:
            return all_objs

        filtered_objects = []
        for name in selected_items:
            item_object = all_objs.find(name=name)
            if item_object is not None:
                filtered_objects.append(item_object)
                selected_items.remove(name)

        for bad_name in selected_items:
            self.logger.warning('"%s" is not a valid profile or system', bad_name)

        if len(filtered_objects) == 0:
            raise ValueError("No valid systems or profiles were specified.")

        return filtered_objects

    def __copy_files(self, iso_distro):
        """
        This method copies the required and optional files from syslinux into the directories we use for building the
        ISO.
        :param iso_distro: The distro (and thus architecture) to build the ISO for.
        """
        self.logger.info("copying syslinux files")

        files_to_copy = [
            "isolinux.bin",
            "menu.c32",
            "chain.c32",
            "ldlinux.c32",
            "libcom32.c32",
            "libutil.c32",
        ]
        optional_files = ["ldlinux.c32", "libcom32.c32", "libutil.c32"]
        syslinux_folders = [
            pathlib.Path(self.api.settings().syslinux_dir),
            pathlib.Path(self.api.settings().syslinux_dir).joinpath("modules/bios/"),
            pathlib.Path("/usr/lib/syslinux/"),
            pathlib.Path("/usr/lib/ISOLINUX/"),
        ]

        # file_copy_success will be used to check for missing files
        file_copy_success: Dict[str, bool] = {
            f: False for f in files_to_copy if f not in optional_files
        }
        for syslinux_folder in syslinux_folders:
            if syslinux_folder.exists():
                for file_to_copy in files_to_copy:
                    source_file = syslinux_folder.joinpath(file_to_copy)
                    if source_file.exists():
                        utils.copyfile(
                            str(source_file),
                            os.path.join(self.isolinuxdir, file_to_copy),
                        )
                        file_copy_success[file_to_copy] = True

        unsuccessful_copied_files = [k for k, v in file_copy_success.items() if not v]
        if len(unsuccessful_copied_files) > 0:
            self.logger.error(
                'The following files were not found: "%s"',
                '", "'.join(unsuccessful_copied_files),
            )
            raise FileNotFoundError(
                "Required file(s) not found. Please check your syslinux installation"
            )

        self.logger.info("copying GRUB2 files")
        bootloader_directory = pathlib.Path(self.api.settings().bootloaders_dir)
        grub_efi = bootloader_directory.joinpath(
            "grub", self.calculate_grub_name(iso_distro)
        )
        buildiso_directory = pathlib.Path(self.api.settings().buildisodir)
        grub_target = buildiso_directory.joinpath("grub", "grub.efi")
        if grub_efi.exists():
            utils.copyfile(str(grub_efi), str(grub_target), symlink=True)
        else:
            self.logger.error('The following files were not found: "%s"', grub_efi)
            raise FileNotFoundError(
                "Required file(s) not found. Please check your GRUB 2 installation"
            )

    def calculate_grub_name(self, distro) -> str:
        """
        This function checks the bootloaders_formats in our settings and then checks if there is a match between the
        architectures and the distribution architecture.
        :param distro: The distribution to get the GRUB2 loader name for.
        """
        loader_formats = self.api.settings().bootloaders_formats
        grub_binary_names: Dict[str, str] = {}
        desired_arch = distro.arch

        for (loader_format, values) in loader_formats.items():
            name = values.get("binary_name", None)
            if name is not None:
                grub_binary_names[loader_format.lower()] = name

        if desired_arch in (Archs.PPC, Archs.PPC64, Archs.PPC64LE, Archs.PPC64EL):
            # GRUB can boot all Power architectures it supports via the following modules directory.
            return grub_binary_names["powerpc-ieee1275"]
        elif desired_arch == Archs.AARCH64:
            # GRUB has only one 64-bit variant it can boot, the name is different how we have named it in Cobbler.
            return grub_binary_names["arm64-efi"]
        elif desired_arch == Archs.ARM:
            # GRUB has only one 32-bit variant it can boot, the name is different how we have named it in Cobbler.
            return grub_binary_names["arm"]

        # Now we do the regular stuff: We map the beginning of the Cobbler arch and try to find suitable loaders.
        # We do want to drop "grub.0" always as it is not efi bootable.
        matches = {
            k: v
            for (k, v) in grub_binary_names.items()
            if k.startswith(desired_arch.value) and v != "grub.0"
        }

        if len(matches) == 0:
            raise ValueError(
                'No matches found for requested Cobbler Arch: "%s"'
                % str(desired_arch.value)
            )
        elif len(matches) == 1:
            return next(iter(matches.values()))
        raise ValueError(
            'Ambiguous matches for GRUB to Cobbler Arch mapping! Requested: "%s" Found: "%s"'
            % (str(desired_arch.value), str(matches.values()))
        )

    def __prepare_buildisodir(self, buildisodir: str = "") -> str:
        """
        This validated the path and type of the buildiso directory and then (re-)creates the apropiate directories.
        :param buildisodir: The directory in which the build of the ISO takes place. If an empty string then the default
                            directory is used.
        :raises ValueError: In case the specified directory does not exist.
        :raises TypeError: In case the specified argument is not of type str.
        :return: The validated and normalized directory with appropriate subfolders provisioned.
        """
        if not isinstance(buildisodir, str):
            raise TypeError("buildisodir needs to be of type str!")
        if not buildisodir:
            buildisodir = self.api.settings().buildisodir
        else:
            if not os.path.isdir(buildisodir):
                raise ValueError("The --tempdir specified is not a directory")

            (_, buildisodir_tail) = os.path.split(os.path.normpath(buildisodir))
            if buildisodir_tail != "buildiso":
                buildisodir = os.path.join(buildisodir, "buildiso")

        self.logger.info('Deleting and recreating the buildisodir at "%s"', buildisodir)
        if os.path.exists(buildisodir):
            shutil.rmtree(buildisodir)
        os.makedirs(buildisodir)

        self.isolinuxdir = os.path.join(buildisodir, "isolinux")

        self.logger.info("building tree for isolinux")
        if not os.path.exists(self.isolinuxdir):
            os.makedirs(self.isolinuxdir)
        self.logger.info("building tree for GRUB 2")
        grub_dir = os.path.join(buildisodir, "grub")
        if not os.path.exists(grub_dir):
            os.makedirs(grub_dir)
        return buildisodir

    def _prepare_iso(
        self,
        buildisodir: str = "",
        iso_distro: str = "",
        profiles: Optional[Union[str, list]] = None,
    ):
        """
        Validates the directories we use for building the ISO and copies files to the right place.
        :param buildisodir: The directory to use for building the ISO. If an empty string then the default directory is
                            used.
        :param iso_distro: The distro to use for building the ISO.
        :param profiles: The profiles to generate the ISO for.
        :return: The normalized directory for further processing.
        """
        try:
            iso_distro = self.api.find_distro(name=iso_distro)
        except ValueError as value_error:
            raise ValueError(
                'Not existent distribution name passed to "cobbler buildiso"!'
            ) from value_error
        buildisodir = self.__prepare_buildisodir(buildisodir)
        self.__copy_files(iso_distro)
        self.profiles = utils.input_string_or_list_no_inherit(profiles)
        return buildisodir

    def _generate_iso(self, xorrisofs_opts: str, iso: str, buildisodir: str):
        """
        Build the final xorrisofs command which is then executed on the disk.
        :param xorrisofs_opts: The additional options for xorrisofs.
        :param iso: The name of the output iso.
        :param buildisodir: The directory in which we build the ISO.
        """
        running_on, _ = utils.os_release()
        if running_on in ("suse", "centos", "virtuozzo", "redhat"):
            isohdpfx_location = pathlib.Path(self.api.settings().syslinux_dir).joinpath(
                "isohdpfx.bin"
            )
        else:
            isohdpfx_location = pathlib.Path(self.api.settings().syslinux_dir).joinpath(
                "mbr/isohdpfx.bin"
            )
        efi_img_location = "grub/grub.efi"
        cmd_list = [
            "xorriso",
            "-as",
            "mkisofs",
            xorrisofs_opts,
            "-isohybrid-mbr",
            str(isohdpfx_location),
            "-c",
            "isolinux/boot.cat",
            "-b",
            "isolinux/isolinux.bin",
            "-no-emul-boot",
            "-boot-load-size",
            "4",
            "-boot-info-table",
            "-eltorito-alt-boot",
            "-e",
            efi_img_location,
            "-no-emul-boot",
            "-isohybrid-gpt-basdat",
            "-V",
            '"COBBLER_INSTALL"',
            "-o",
            iso,
            buildisodir,
        ]
        cmd = " ".join(cmd_list)

        xorrisofs_return_code = utils.subprocess_call(cmd, shell=True)
        if xorrisofs_return_code != 0:
            self.logger.error("xorrisofs failed with non zero exit code!")
            return

        self.logger.info("ISO build complete")
        self.logger.info("You may wish to delete: %s", buildisodir)
        self.logger.info("The output file is: %s", iso)
